from collections import namedtuple
from copy import deepcopy

import numpy as np
import scipy
import sklearn as skl
from scikit_obliquetree.basic_functions import Struct, piter
from sklearn.base import BaseEstimator
from sklearn.metrics import accuracy_score as accuracy
from sklearn.tree import DecisionTreeRegressor


class GradientBoosting(BaseEstimator):
    """Realization of gradient boosting predictor.
    Params:
        base_learner (object): instance of base learner class, initialized with necessary parameters.
        All base learners inside are recreated with these parameters.
        loss (string): loss function. Possible values are: "square", "exp" and "log".
        base_learners_count (int): how many base learners to fit (number of boosting iterations)
        fit_coefs (boolean): whether to fit or not multiplier coefficients by each base learner
        refit_tree (boolean): in case the base learner is regression tree, whether to refit or not leaf predictions of each tree.
        shrinkage (float): how much to multiply each coefficient by the base learner
        log_level (int): how many debug messages to display. The lower the value, the more logger messages will be shown.
    Comments:
        - Univarite regression predictions are made for loss="square".
        - Binary classification predictions are made for loss="exp" or loss="log". In these cases y=0 or y=1.
        - For loss="log" not only classes can be predicted (with predict function) but also class probabilities (with predict_proba function)
        - Zero-th approximation is zero. Higher order approximations are sums of base learners with coefficients.
    Author:
        Victor Kitov, 03.2016."""

    def __init__(
        self,
        base_learner,
        base_learners_count,
        loss_name="square",
        fit_coefs=True,
        refit_tree=True,
        shrinkage=0.1,
        max_fun_evals=200,
        xtol=10 ** -6,
        ftol=1e-6,
    ):
        self.base_learners_count = base_learners_count
        self.base_learner = base_learner
        self.fit_coefs = fit_coefs
        self.shrinkage = shrinkage
        self.refit_tree = refit_tree
        self.optimization = Struct(
            max_fun_evals=max_fun_evals, xtol=xtol, ftol=ftol
        )

        self.coefs = []
        self.base_learners = []
        self.loss_name = loss_name
        self.max_fun_evals = max_fun_evals
        self.xtol = xtol
        self.ftol = ftol

        if loss_name == "square":
            self.loss = lambda r, y: 0.5 * (r - y) ** 2
            self.loss_derivative = lambda r, y: (r - y)
            self.task = "regression"
        elif loss_name == "exp":
            self.loss = lambda r, y: np.exp(-r * y)
            self.loss_derivative = lambda r, y: -(y * np.exp(-r * y))
            self.task = "classification"
        elif loss_name == "log":
            self.loss = lambda r, y: np.log(1 + np.exp(-r * y))
            self.loss_derivative = lambda r, y: -(y / (1 + np.exp(r * y)))
            self.task = "classification"
        else:
            raise Exception('Not implemented loss "%s"' % loss_name)

    def fit(
        self,
        X,
        y,
        X_val=None,
        y_val=None,
        bad_iters_count=+np.inf,
        cross_val=False,
    ):
        """If called like fit(X,y), then the number of base_learners is always base_learners_count (specified at initialization).
        If called like fit(self, X, y, X_val, y_val, bad_iters_count) at most there are also base_learners_count base learners but may be less due to
        early stopping:
        At each iteration accuracy (using validation set, specified by X_val [design matrix], y_val [outputs]) is estimated and
        position of best iteration tracked. If there were >=bad_iters_count after the best iteration, fitting process stops."""

        X = X.astype("float32")
        y = y.astype("float32")
        D = X.shape[1]
        min_loss = +np.inf
        min_pos = -1
        fit_after = False

        if self.task == "classification":
            assert all(np.unique(y) == [0, 1]), "Only y=0 or y=1 supported!"
            y[y == 0] = -1  # inner format of classes y=+1 or y=-1
        N = len(X)
        F_current = self.F(X)  # current value, all zeros if not tuned before.

        if cross_val:
            self.base_learners_after = []
            self.coefs_after = []

        if X_val != None and y_val != None:
            self.loss_val = []
            F_val = self.F(X_val)

        for iter_num in piter(
            range(self.base_learners_count), percent_period=3, show=False
        ):

            if X_val != None and y_val != None:
                if self.task == "regression":
                    Y_val_hat = F_val
                    loss = mean(
                        abs(Y_val_hat - y_val)
                    )  # MAE tracking on validation set
                else:  # classification
                    Y_val_hat = (F_val >= 0).astype(int)
                    loss = 1 - skl.metrics.accuracy_score(y_val, Y_val_hat)
                self.loss_val.append(loss)
                if not fit_after:
                    if loss < min_loss:
                        min_pos = iter_num
                        min_loss = loss

                    if iter_num - min_pos >= bad_iters_count:
                        self.base_learners_count = len(self.base_learners)
                        if cross_val:
                            fit_after = True
                        else:
                            break

            z = -self.loss_derivative(F_current, y)

            base_learner = deepcopy(self.base_learner)  # recreate base learner
            base_learner.fit(X, z)

            if isinstance(base_learner, DecisionTreeRegressor) and (
                self.refit_tree == True
            ):  # tree refitting
                leaf_ids = base_learner.tree_.apply(X)
                unique_leaf_ids = np.unique(leaf_ids)
                for leaf_id in unique_leaf_ids:
                    leaf_pos_sels = leaf_ids == leaf_id
                    prediction = base_learner.tree_.value[leaf_id, 0, 0]

                    def loss_at_leaf(value):
                        return np.sum(
                            self.loss(
                                F_current[leaf_pos_sels] + value,
                                y[leaf_pos_sels],
                            )
                        )

                    refined_prediction = scipy.optimize.fmin(
                        loss_at_leaf,
                        prediction,
                        xtol=self.optimization.xtol,
                        ftol=self.optimization.ftol,
                        maxfun=self.optimization.max_fun_evals,
                        disp=0,
                    )

                    base_learner.tree_.value[leaf_id, 0, 0] = refined_prediction

            base_pred = base_learner.predict(X)

            if (
                self.fit_coefs == False
            ):  # coefficients by base learner refitting
                coef = 1.0 / self.base_learners_count
            else:

                def loss_after_weighted_addition(coef):
                    return np.sum(self.loss(F_current + coef * base_pred, y))

                res = scipy.optimize.fmin(
                    loss_after_weighted_addition,
                    1,
                    xtol=self.optimization.xtol,
                    ftol=self.optimization.ftol,
                    maxfun=self.optimization.max_fun_evals,
                    disp=0,
                )
                coef = res[0]
                # if coef<0:
                #    self.log.pr3('coef=%s is negative!' % coef)
                # if coef==0:
                #    self.log.pr3('coef=%s is zero!' % coef)

            coef = coef * self.shrinkage

            if fit_after:
                self.coefs_after.append(coef)
                self.base_learners_after.append(base_learner)
            else:
                if cross_val:
                    self.coefs_after.append(coef)
                    self.base_learners_after.append(base_learner)
                self.coefs.append(coef)
                self.base_learners.append(base_learner)

            F_current += coef * base_pred
            if X_val != None and y_val != None:
                F_val += coef * base_learner.predict(X_val)

    def F(self, X, max_base_learners_count=np.inf):
        """Internal function used for forecasting.
        X-design matrix, each row is an object for which a forecast should be made.
        max_base_learners_count - maximal iteration at which to stop. F is evaluated for min(max_base_learners_count, len(self.base_learners)) models."""

        F_val = np.zeros(len(X))

        for iter_num, (coef, base_learner) in enumerate(
            zip(self.coefs, self.base_learners)
        ):
            base_pred = base_learner.predict(X)
            F_val += coef * base_pred
            if iter_num + 1 >= max_base_learners_count:
                break

        return F_val

    def predict(self, X, base_learners_count=np.inf):
        if self.task == "regression":
            return self.F(X, base_learners_count)
        else:  # classification
            return (self.F(X, base_learners_count) >= 0).astype(
                int
            )  # F(X)>=0 = > predition=1 otherwise prediction=0

    def predict_proba(self, X, base_learners_count=np.inf):
        """Predict class probabilities for objects, specified by rows of matrix X.
        iter_num - at what iteration to stop. If not specified all base learners are used.
        Applicable only for loss function="log". Classes are stored in self.classes_ attribute."""

        if self.loss != "log":
            raise Exception("Inapliccable for loss %s" % self.loss)
        self.classes_ = [0, 1]
        scores = self.F(X, base_learners_count)
        probs = 1 / (1 + exp(-scores))
        return hstack([1 - probs, probs])

    def get_losses(self, X, Y):
        """Estimate loss for each iteration of the prediction process of boosting.
        Returns an array losses with length=#{base learners}"""

        losses = np.zeros(len(self.base_learners))
        F_val = np.zeros(len(X))

        for iter_num, (coef, base_learner) in enumerate(
            zip(self.coefs, self.base_learners)
        ):
            base_pred = base_learner.predict(X)
            F_val += coef * base_pred

            if self.task == "regression":
                Y_hat = F_val
                losses[iter_num] = mean(abs(Y_hat - Y)) / mean(abs(Y))
            else:  # classification
                Y_hat = (F_val >= 0).astype(int)
                losses[iter_num] = 1 - skl.metrics.accuracy_score(Y, Y_hat)

        return losses

    def score(self, X, y):
        y_pred = predict(X)
        return accuracy(y_pred, y)

    def remove_redundant_base_learners(self, X_val, Y_val, max_count=+np.inf):
        """Using validation set, specified by (X_val, y_val) find optimal number of base learners
        (at this number the loss on validation is minimal).
        All base learners above this number are removed.
        max_count - is the maximum possible number of base learners retained."""

        orig_count = len(self.base_learners)
        losses = self.get_losses(X_val, Y_val)
        self.base_learners_count = min(argmin(losses) + 1, max_count)
        self.coefs = self.coefs[: self.base_learners_count]
        self.base_learners = self.base_learners[: self.base_learners_count]
        self.log.pr1(
            "Cut base learners count from %d to %d."
            % (orig_count, self.base_learners_count)
        )



from copy import deepcopy

import numpy as np
from scipy.linalg import norm
from sklearn.base import BaseEstimator
from sklearn.cluster import KMeans
from sklearn.linear_model import LogisticRegression


class Node:
    def __init__(self, objects, labels, **kwargs):

        self.objects = objects
        self.labels = labels
        self.centroid = objects.mean(axis=0)
        self.class_ = kwargs.get("class_", None)
        self.is_leaf = kwargs.get("is_leaf", False)
        self._weights = kwargs.get("weights", None)
        self._left_child = kwargs.get("left_child", None)
        self._right_child = kwargs.get("right_child", None)
        self._features_indices = kwargs.get("indices", None)

    def get_child(self, datum):
        if self.is_leaf:
            raise Exception("Leaf node does not have children.")
        X = deepcopy(datum)
        X = datum[self._features_indices]

        if X.dot(np.array(self._weights[:-1]).T) + self._weights[-1] < 0:
            return self.left_child
        else:
            return self.right_child

    @property
    def left_child(self):
        if self.is_leaf:
            raise Exception("Leaf node does not have split rule.")
        return self._left_child

    @property
    def right_child(self):
        if self.is_leaf:
            raise Exception("Leaf node does not have split rule.")
        return self._right_child


class Clustering(BaseEstimator):
    def __init__(self, method="k-means", n_clusters=6):
        self.method = method
        self.n_clusters = n_clusters

    def fit(self, X):
        if self.method == "k-means":
            clf = KMeans(n_clusters=self.n_clusters)
        else:
            raise Exception
        clf.fit(X)
        y = clf.predict(X)

        self.clusters_ = []
        for cluster in np.unique(y):
            self.clusters_.append(np.where(y == cluster)[0])

        self.clusters_ = np.array(self.clusters_, dtype=object)


class BUTIF(BaseEstimator):
    def __init__(
        self,
        clustering_method="k-means",
        max_leaf=5,
        linear_model=LogisticRegression(),
        task="classification",
        selector=None,
        best_k=None,
    ):
        self.clustering_method = clustering_method
        self.max_leaf = max_leaf
        self.clustering = Clustering(
            method=self.clustering_method, n_clusters=self.max_leaf
        )
        self.linear_model = linear_model
        self.selector = selector
        self._root = None
        self._nodes = []
        self._leaves = {}
        self.best_k = best_k
        self._k_features = best_k
        self.task = task

    def fit(self, X, y):
        self.clustering.n_clusters = min(
            self.clustering.n_clusters, len(np.unique(y))
        )

        def merging(L, classes_):
            if len(L) == 1:
                single_node = list(L.keys())[0]
                return L[single_node]
            else:
                dist_min = np.inf
                distance = 0
                for i in L.keys():
                    for j in L.keys():
                        if i == j:
                            continue
                        if (
                            self.task == "classification"
                            and L[i].class_ == L[j].class_
                        ):
                            continue

                        distance = norm(L[i].centroid - L[j].centroid)

                        if distance < dist_min:
                            dist_min = distance
                            i_node, j_node = i, j
                            left_child = L[i]
                            right_child = L[j]

                objects_ = np.vstack((left_child.objects, right_child.objects))
                labels_ = np.ones(objects_.shape[0])
                labels_[: left_child.objects.shape[0]] = -1.0
                labels_[left_child.objects.shape[0] :] = 1.0

                indices_ = np.arange(n_features)
                if self.selector is not None:
                    self.selector.fit(objects_, labels_)
                    indices_ = self.selector.scores_.argsort()[
                        : self._k_features
                    ]

                self.linear_model.fit(objects_[:, indices_], labels_)
                weights_ = np.zeros(len(indices_) + 1)
                weights_[:-1] = self.linear_model.coef_
                weights_[-1] = self.linear_model.intercept_

                labels_[: left_child.objects.shape[0]] = left_child.labels
                labels_[left_child.objects.shape[0] :] = right_child.labels

                new_node = Node(
                    objects_,
                    labels_,
                    left_child=left_child,
                    right_child=right_child,
                    class_=classes_ + 1,  # new meta class
                    indices=indices_,
                    weights=weights_,
                )
                del L[i_node]
                del L[j_node]
                L[i_node] = new_node
                return merging(L, classes_ + 1)

        n_objects, n_features = X.shape

        leaf_cnt = 0
        if self.task == "classification":
            self.classes_ = np.unique(y)
            for c in self.classes_:
                objects_in_c = X[y == c]

                self.clustering.fit(objects_in_c)
                partition = self.clustering.clusters_

                for cluster in partition:
                    leaf = Node(
                        objects_in_c[cluster],
                        c * np.ones(len(cluster)),
                        is_leaf=True,
                    )
                    leaf.class_ = c
                    self._leaves[str(leaf_cnt)] = leaf
                    leaf_cnt += 1
        else:
            # self.clustering.fit(X)
            self.clustering.fit(y.reshape(-1, 1))
            partition = self.clustering.clusters_
            for i, cluster in enumerate(partition):
                leaf = Node(
                    X[cluster.astype(int)],
                    i * np.ones(len(cluster)),
                    is_leaf=True,
                )
                leaf.class_ = np.mean(y[cluster.astype(int)])
                self._leaves[str(leaf_cnt)] = leaf
                leaf_cnt += 1
            self.classes_ = np.arange(len(partition))

        self._root = merging(self._leaves, len(self.classes_))

    def predict(self, X):
        def _predict_single(X):
            cur_node = self._root
            while not cur_node.is_leaf:
                cur_node = cur_node.get_child(X)
            return cur_node.class_

        if self._root is None:
            raise Exception("Decision tree has not been trained.")
        n_objects = X.shape[0]
        predictions = np.zeros((n_objects,), dtype=float)
        for i in range(n_objects):
            predictions[i] = _predict_single(X[i])
        return predictions



from abc import ABCMeta, abstractmethod

import numpy as np


class MSE:
    """
    Mean squared error impurity criterion
    """

    def __call__(self, left_label, right_label):
        left_len, right_len = len(left_label), len(right_label)

        left_std = np.std(left_label)
        right_std = np.std(right_label)

        total = left_len + right_len

        return (left_len / total) * left_std + (right_len / total) * right_std


class SegmentorBase:
    """
    Abstract segmentor class. Segmentor called in nodes for find best split.


    Parameters
    -----------

    msl : int, optional (default=1)
    The minimum number of samples required to be at a leaf nodes.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def _split_generator(self, X):
        """
        Abstract method for split.


        Parameter
        -----------

        X : array-like, shape = [n_samples, n_features]
        The training input samples.
        """
        pass

    def __init__(self, msl=1):
        self._min_samples_leaf = msl

    def __call__(self, X, y, impurity=MSE()):
        """
        Parameters
        -----------

        X : array-like, shape = [n_samples, n_features]
        The training input samples.
        y : array-like, shape = [n_samples]
        The target values.
        impurity : object of impurity (default=MSE()).
        The name of criterion to measure the quality of a split.


        Returns
        -----------

        Tuple of the following elements:
        best_impurity : float.
        The best value of impurity.
        best_split_rule : tuple.
        The pair of feature and value.
        best_left_i : numpy.ndarray.
        The indexes of left node objects.
        best_right_i : numpy.ndarray.
        The indexes of right node objects.
        """
        best_impurity = float("inf")
        best_split_rule = None
        best_left_i = None
        best_right_i = None
        splits = self._split_generator(X)

        for left_i, right_i, split_rule in splits:
            if (
                left_i.size >= self._min_samples_leaf
                and right_i.size >= self._min_samples_leaf
            ):
                left_labels, right_labels = y[left_i], y[right_i]
                cur_impurity = impurity(left_labels, right_labels)
                if cur_impurity < best_impurity:
                    best_impurity = cur_impurity
                    best_split_rule = split_rule
                    best_left_i = left_i
                    best_right_i = right_i
        return (best_impurity, best_split_rule, best_left_i, best_right_i)


class MeanSegmentor(SegmentorBase):
    """
    Split based on mean value of each feature.
    """

    def _split_generator(self, X):
        """
        Parameters
        -----------

        X : array-like, shape = [n_samples, n_features]
        The training input samples.


        Returns
        -----------

        Tuple of the following elements:
        left_i : numpy.ndarray.
        The indexes of left node objects.
        right_i : numpy.ndarray.
        The indexes of right node objects.
        split_rule : tuple.
        The pair of feature and value.
        """
        for feature_i in range(X.shape[1]):
            feature_values = X[:, feature_i]
            mean = np.mean(feature_values)
            left_i = np.nonzero(feature_values < mean)[0]
            right_i = np.nonzero(feature_values >= mean)[0]
            split_rule = (feature_i, mean)
            yield (left_i, right_i, split_rule)



from copy import deepcopy

import numpy as np
from numba import njit
from numpy.linalg import norm
from scikit_obliquetree.HHCART import HHCARTNode
from sklearn.base import BaseEstimator


@njit(cache=True)
def np_apply_along_axis(func1d, axis, arr):
    assert arr.ndim == 2
    assert axis in [0, 1]
    if axis == 0:
        result = np.empty(arr.shape[1])
        for i in range(len(result)):
            result[i] = func1d(arr[:, i])
    else:
        result = np.empty(arr.shape[0])
        for i in range(len(result)):
            result[i] = func1d(arr[i, :])
    return result


@njit(cache=True)
def np_max(array, axis):
    return np_apply_along_axis(np.max, axis, array)


@njit(cache=True)
def objective(w, theta_left, theta_right, X, y):
    loss = np.vstack(
        (
            -w.dot(X.T) + (theta_left - y) ** 2,
            w.dot(X.T) + (theta_right - y) ** 2,
        )
    )
    return (np_max(loss, axis=0) - np.abs(w.dot(X.T))).sum()


@njit(cache=True)
def optimization(
    X_train,
    y,
    left_indices,
    right_indices,
    sr,
    tol,
    thau,
    step,
    nu,
    max_iter,
    converge_criterion,
):
    n_objects, n_features = X_train.shape
    w_new = np.zeros(n_features)
    w_new[sr[0]] = 1.0
    w_new[-1] = -sr[1]
    theta_left = y[left_indices].mean()
    theta_right = y[right_indices].mean()
    l = []
    l_old = np.inf
    l_new = objective(w_new, theta_left, theta_right, X_train, y)
    it = 0
    while True:
        if converge_criterion == "zero_loss" and (np.abs(l_new) <= tol):
            break
        if converge_criterion == "no_improvement" and (
            np.abs(l_old - l_new) <= tol
        ):
            break

        it += 1
        w_old = w_new
        for i in range(thau):
            ind_batch = np.random.choice(np.arange(n_objects), size=1)
            X_batch = X_train[ind_batch, :]
            y_batch = y[ind_batch]

            s = np.sign((w_old * X_batch).sum())

            if (
                -(w_new * X_batch).sum() + (theta_left - y_batch) ** 2
                >= (w_new * X_batch).sum() + (theta_right - y_batch) ** 2
            ):
                w_new = w_new + (step * (1 + s) * X_batch).flatten()
                theta_left = (theta_left - 2 * step * (theta_left - y_batch))[0]
            else:
                w_new = w_new - (step * (1 - s) * X_batch).flatten()
                theta_right = (
                    theta_right - 2 * step * (theta_right - y_batch)
                )[0]

            if (w_new * w_new).sum() > nu:
                w_new = np.sqrt(thau) * w_new / np.linalg.norm(w_new)

        l_old = objective(w_old, theta_left, theta_right, X_train, y)
        l_new = objective(w_new, theta_left, theta_right, X_train, y)
        l.append(objective(w_new, theta_left, theta_right, X_train, y))
        if it >= max_iter:
            break
    return theta_left, theta_right, w_new


class Node(HHCARTNode):
    def get_child(self, datum):
        if self.is_leaf:
            raise Exception("Leaf node does not have children.")
        X = deepcopy(datum)

        if X.dot(np.array(self._weights).T) >= 0:
            return self.left_child
        else:
            return self.right_child


class ContinuouslyOptimizedObliqueRegressionTree(BaseEstimator):
    def __init__(
        self,
        impurity,
        segmentor,
        nu=1.0,
        thau=10,
        max_iter=100,
        tol=1e-6,
        step=0.1,
        converge_criterion="no_improvement",
        **kwargs,
    ):
        self.impurity = impurity
        self.segmentor = segmentor
        self.nu = nu
        self.thau = thau
        self.max_iter = max_iter
        self.tol = tol
        self.step = step
        self.converge_criterion = converge_criterion
        self._max_depth = kwargs.get("max_depth", None)
        self._min_samples = kwargs.get("min_samples", 2)
        self._root = None
        self._nodes = []
        assert converge_criterion in ["zero_loss", "no_improvement"]

    def fit(self, X, y):
        self._root = self._generate_node(X, y, 0)

    def _terminate(self, X, y, cur_depth):
        if self._max_depth != None and cur_depth == self._max_depth:
            # maximum depth reached.
            return True
        elif y.size < self._min_samples:
            # minimum number of samples reached.
            return True
        elif np.unique(y).size == 1:
            return True
        else:
            return False

    def _generate_leaf_node(self, cur_depth, y):
        node = Node(cur_depth, y, is_leaf=True)
        self._nodes.append(node)
        return node

    def _generate_node(self, X, y, cur_depth):
        if self._terminate(X, y, cur_depth):
            return self._generate_leaf_node(cur_depth, y)
        else:
            impurity, sr, left_indices, right_indices = self.segmentor(
                X, y, self.impurity
            )

            if not sr:
                return self._generate_leaf_node(cur_depth, y)

            X_stack = np.hstack((X, np.ones(X.shape[0])[:, np.newaxis]))

            theta_left, theta_right, w_new = optimization(
                X_stack,
                y,
                left_indices,
                right_indices,
                sr,
                self.tol,
                self.thau,
                self.step,
                self.nu,
                self.max_iter,
                self.converge_criterion,
            )

            weights = w_new
            assert len(weights) == (
                X.shape[1] + 1
            ), f"{len(weights), (X.shape[1] + 1)}"

            # generate indices
            mask = (weights.dot(X_stack.T) >= 0).flatten()
            left_indices = np.arange(0, len(X))[mask]
            right_indices = np.arange(0, len(X))[np.logical_not(mask)]

            if np.sum(mask) == 0 or np.sum(np.logical_not(mask)) == 0:
                return self._generate_leaf_node(cur_depth, y)

            X_left, y_left = X[left_indices], y[left_indices]
            X_right, y_right = X[right_indices], y[right_indices]

            node = Node(
                cur_depth,
                y,
                split_rules=sr,
                weights=weights,
                left_child=self._generate_node(X_left, y_left, cur_depth + 1),
                right_child=self._generate_node(
                    X_right, y_right, cur_depth + 1
                ),
                is_leaf=False,
            )
            self._nodes.append(node)
            return node

    def predict(self, X):
        X = np.hstack((X, np.ones(X.shape[0])[:, np.newaxis]))

        def predict_single(datum):
            cur_node = self._root
            while not cur_node.is_leaf:
                cur_node = cur_node.get_child(datum)
            return cur_node.label

        if not self._root:
            raise Exception("Decision tree has not been trained.")
        size = X.shape[0]
        predictions = np.empty((size,), dtype=float)
        for i in range(size):
            predictions[i] = predict_single(X[i, :])
        return predictions



def piter(x, percent_period=1, period=None, end="| ", show=True):
    """Iterates through x (any iterable object, having len) returning iteratively elements from x and printing progress.
    Progress is printed every <period> iterations or after every <percent_period> percent of total was complete.
    Useful for controlling how much of the long computations were completed.
    Example:
        for i in piter([10,11,12,13,14,15],2):
            print(i)
    Output:
        0.00% done
        10
        11
        33.33% done
        12
        13
        66.67% done
        14
        15
    Author: Victor Kitov (v.v.kitov@yandex.ru), 03.2016."""

    if show == False:  # do nothing
        for element in x:
            yield element
    else:

        if hasattr(x, "__len__"):
            for i, element in enumerate(x):
                yield element
        else:
            for i, element in enumerate(x):
                yield element


class Struct:
    """
    Structure data type.
    Examples of use:
    A=cStruct()
    A.property1=value1
    A.property2=value2
    ...
    A=cStruct(property1=value1,property2=value2,...)
    """

    def __init__(self, **keywords):
        self._fields = []
        for sKey in list(keywords.keys()):
            setattr(self, sKey, keywords[sKey])

    def get_str(self, sSeparator):
        sString = ""
        lsAttributes = list(vars(self).keys())

        for sAttribute in lsAttributes:
            if sAttribute[0] != "_":
                attr = getattr(self, sAttribute)
                if isinstance(attr, (np.ndarray, pd.Series, pd.DataFrame)):
                    sString += f"{sAttribute}=...,{sSeparator}"
                else:
                    sAttributeValue = str(getattr(self, sAttribute))
                    if len(sAttributeValue) > 30:
                        sString += f"{sAttribute}=...,{sSeparator}"
                    else:
                        sString += (
                            f"{sAttribute}={sAttributeValue},{sSeparator}"
                        )
        return "{" + sString[:-2] + "}"

    @property
    def pstr(self):
        return self.get_str("\n")

    def __str__(self):
        return self.get_str(" ")

    def __repr__(self):
        return self.get_str(" ")

    def __unicode__(self):
        return self.get_str(" ")

    def __eq__(self, other):
        return (
            self.GetAttributes2ValuesDict() == other.GetAttributes2ValuesDict()
        )

    def __ne__(self, other):
        return (
            self.GetAttributes2ValuesDict() != other.GetAttributes2ValuesDict()
        )

    def __hash__(self):
        return hash(self.GetAttributes2ValuesTuple())

    def get_defaults(self, oDefaultStruct):
        """
        A=cStruct(i=1,j=2,k=3)
        B=cStruct(a=10,b=20,c=30,i=333)
        A.GetDefaults(B)
        print A
        Structure: a=10 c=30 b=20 i=1 k=3 j=2"""
        lsDefaultAttributes = list(vars(oDefaultStruct).keys())
        for sDefaultAttribute in lsDefaultAttributes:
            if not hasattr(self, sDefaultAttribute):
                setattr(
                    self,
                    sDefaultAttribute,
                    getattr(oDefaultStruct, sDefaultAttribute),
                )

    @property
    def fields(self):
        return self._fields

    def get_fields(self):
        return self._fields

    @property
    def fields2values(self):
        dAttributes2Values = {}
        for sAttribute in list(vars(self).keys()):
            dAttributes2Values[sAttribute] = getattr(self, sAttribute)
        return dAttributes2Values

    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        if name != "_fields":
            if name not in self._fields:
                self._fields.append(name)

    def __delattr__(self, name):
        super().__delattr__(name)
        if name != "_fields":
            if name in self._fields:
                self._fields.remove(name)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scikit_obliquetree']

package_data = \
{'': ['*']}

install_requires = \
['rich>=9.8.2,<10.0.0', 'typer[all]>=0.3.2,<0.4.0']

extras_require = \
{':python_version < "3.8"': ['importlib_metadata>=1.6.0,<2.0.0']}

entry_points = \
{'console_scripts': ['scikit-obliquetree = scikit_obliquetree.__main__:app']}

setup_kwargs = {
    'name': 'scikit-obliquetree',
    'version': '0.1.4',
    'description': 'Oblique Decision Tree in Python',
    'long_description': '# scikit-obliquetree\n\n<div align="center">\n\n[![Build status](https://github.com/zhenlingcn/scikit-obliquetree/workflows/build/badge.svg?branch=master&event=push)](https://github.com/zhenlingcn/scikit-obliquetree/actions?query=workflow%3Abuild)\n[![Python Version](https://img.shields.io/pypi/pyversions/scikit-obliquetree.svg)](https://pypi.org/project/scikit-obliquetree/)\n[![Dependencies Status](https://img.shields.io/badge/dependencies-up%20to%20date-brightgreen.svg)](https://github.com/zhenlingcn/scikit-obliquetree/pulls?utf8=%E2%9C%93&q=is%3Apr%20author%3Aapp%2Fdependabot)\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Security: bandit](https://img.shields.io/badge/security-bandit-green.svg)](https://github.com/PyCQA/bandit)\n[![Pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/zhenlingcn/scikit-obliquetree/blob/master/.pre-commit-config.yaml)\n[![Semantic Versions](https://img.shields.io/badge/%F0%9F%9A%80-semantic%20versions-informational.svg)](https://github.com/zhenlingcn/scikit-obliquetree/releases)\n[![License](https://img.shields.io/github/license/zhenlingcn/scikit-obliquetree)](https://github.com/zhenlingcn/scikit-obliquetree/blob/master/LICENSE)\n\nOblique Decision Tree in Python\n\n</div>\n\n## Introduction\n\nThe oblique decision tree is a popular choice in the machine learning domain for improving the performance of traditional decision tree algorithms. In contrast to the traditional decision tree, which uses an axis-parallel split point to determine whether a data point should be assigned to the left or right branch of a decision tree, the oblique decision tree uses a hyper-plane based on all data point features. \n\nNumerous works in the machine learning domain have shown that oblique decision trees can achieve exceptional performance in a wide range of domains. However, there is still a lack of a package that has implemented oblique decision tree algorithms, which stymies the development of this domain. As a result, the goal of this project is to solve this problem by implementing some well-known algorithms in this domain. We hope that by doing so, these algorithms will serve as a baseline for machine learning practitioners to compare newly designed algorithms to existing algorithms.\n\n\n## 🚀 Features\n* A simple scikit-learn interface for oblique decision tree algorithms\n* A general gradient boosting estimator that can be used to improve arbitrary base estimators\n\n## Installation\n\n```bash\npip install -U scikit-obliquetree\n```\n\nor install with `Poetry`\n\n```bash\npoetry add scikit-obliquetree\n```\n\nThen you can run\n\n```bash\nscikit-obliquetree --help\n```\n\n```bash\nscikit-obliquetree --name Roman\n```\n\nor if installed with `Poetry`:\n\n```bash\npoetry run scikit-obliquetree --help\n```\n\n```bash\npoetry run scikit-obliquetree --name Roman\n```\n\n## Example\nExample of usage:\n```python\nfrom sklearn.datasets import load_boston\nfrom sklearn.ensemble import BaggingRegressor\nfrom sklearn.model_selection import cross_val_score\n\nfrom scikit_obliquetree.HHCART import HouseHolderCART\nfrom scikit_obliquetree.segmentor import MSE, MeanSegmentor\n\nX, y = load_boston(return_X_y=True)\nreg = BaggingRegressor(\n    HouseHolderCART(MSE(), MeanSegmentor(), max_depth=3),\n    n_estimators=100,\n    n_jobs=-1,\n)\nprint(\'CV Score\', cross_val_score(reg, X, y))\n```\n\n## 🛡 License\n\n[![License](https://img.shields.io/github/license/zhenlingcn/scikit-obliquetree)](https://github.com/zhenlingcn/scikit-obliquetree/blob/master/LICENSE)\n\nThis project is licensed under the terms of the `Apache Software License 2.0` license. See [LICENSE](https://github.com/zhenlingcn/scikit-obliquetree/blob/master/LICENSE) for more details.\n\n## 📃 Citation\n\n```\n@misc{scikit-obliquetree,\n  author = {ECNU},\n  title = {Oblique Decision Tree in Python},\n  year = {2021},\n  publisher = {GitHub},\n  journal = {GitHub repository},\n  howpublished = {\\url{https://github.com/zhenlingcn/scikit-obliquetree}}\n}\n```\n\n## Credits\n\nThis project was generated with [`python-package-template`](https://github.com/TezRomacH/python-package-template).\n',
    'author': 'ECNU',
    'author_email': 'zhenlingcn@foxmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/zhenlingcn/scikit-obliquetree',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from copy import deepcopy

import numpy as np
from scipy.linalg import norm
from sklearn.base import BaseEstimator
from sklearn.decomposition import PCA
from sklearn.linear_model import Ridge


class HHCARTNode:
    def __init__(self, depth, labels, **kwargs):
        self.depth = depth
        self.labels = labels
        self.is_leaf = kwargs.get("is_leaf", False)
        self._split_rules = kwargs.get("split_rules", None)
        self._weights = kwargs.get("weights", None)
        self._left_child = kwargs.get("left_child", None)
        self._right_child = kwargs.get("right_child", None)

        if not self.is_leaf:
            assert self._split_rules
            assert self._left_child
            assert self._right_child

    def get_child(self, datum):
        if self.is_leaf:
            raise Exception("Leaf node does not have children.")
        X = deepcopy(datum)

        if X.dot(np.array(self._weights[:-1]).T) - self._weights[-1] < 0:
            return self.left_child
        else:
            return self.right_child

    @property
    def label(self):
        if not hasattr(self, "_label"):
            self._label = np.mean(self.labels)
        return self._label

    @property
    def split_rules(self):
        if self.is_leaf:
            raise Exception("Leaf node does not have split rule.")
        return self._split_rules

    @property
    def left_child(self):
        if self.is_leaf:
            raise Exception("Leaf node does not have split rule.")
        return self._left_child

    @property
    def right_child(self):
        if self.is_leaf:
            raise Exception("Leaf node does not have split rule.")
        return self._right_child


class HouseHolderCART(BaseEstimator):
    def __init__(self, impurity, segmentor, method="eig", tau=1e-4, **kwargs):
        self.impurity = impurity
        self.segmentor = segmentor
        self.method = method
        self.tau = tau
        self._max_depth = kwargs.get("max_depth", None)
        self._min_samples = kwargs.get("min_samples", 2)
        self._alpha = kwargs.get("alpha", None)  # only for linreg method
        self._root = None
        self._nodes = []

    def _terminate(self, X, y, cur_depth):
        if self._max_depth != None and cur_depth == self._max_depth:
            # maximum depth reached.
            return True
        elif y.size < self._min_samples:
            # minimum number of samples reached.
            return True
        elif np.unique(y).size == 1 or np.unique(X, axis=0).shape[0] == 1:
            return True
        else:
            return False

    def _generate_leaf_node(self, cur_depth, y):
        node = HHCARTNode(cur_depth, y, is_leaf=True)
        self._nodes.append(node)
        return node

    def _generate_node(self, X, y, cur_depth):
        if self._terminate(X, y, cur_depth):
            return self._generate_leaf_node(cur_depth, y)
        else:
            n_objects, n_features = X.shape
            impurity_best, sr, left_indices, right_indices = self.segmentor(
                X, y, self.impurity
            )
            oblique_split = False

            # generate Housholder matrix
            if self.method == "eig":
                extractor = PCA(n_components=1)
                extractor.fit(X, y)
                mu = extractor.components_[0]
            if self.method == "ridge":
                reg_model = Ridge(alpha=self._alpha)
                reg_model.fit(X, y)
                mu = reg_model.coef_

            I = np.diag(np.ones(n_features))
            check_ = np.sqrt(((I - mu) ** 2).sum(axis=1))
            if (check_ > self.tau).sum() > 0:
                i = np.argmax(check_)
                e = np.zeros(n_features)
                e[i] = 1.0
                w = (e - mu) / norm(e - mu)
                householder_matrix = I - 2 * w[:, np.newaxis].dot(
                    w[:, np.newaxis].T
                )
                X_house = X.dot(householder_matrix)
                (
                    impurity_house,
                    sr_house,
                    left_indices_house,
                    right_indices_house,
                ) = self.segmentor(X_house, y, self.impurity)

                if impurity_best > impurity_house:
                    oblique_split = True
                    impurity_best = impurity_house
                    left_indices = left_indices_house
                    right_indices = right_indices_house
                    sr = sr_house
            else:
                householder_matrix = I

            if not sr:
                return self._generate_leaf_node(cur_depth, y)

            i, treshold = sr
            weights = np.zeros(n_features + 1)
            if oblique_split:
                weights[:-1] = householder_matrix[:, i]
            else:
                weights[i] = 1
            weights[-1] = treshold

            X_left, y_left = X[left_indices], y[left_indices]
            X_right, y_right = X[right_indices], y[right_indices]

            node = HHCARTNode(
                cur_depth,
                y,
                split_rules=sr,
                weights=weights,
                left_child=self._generate_node(X_left, y_left, cur_depth + 1),
                right_child=self._generate_node(
                    X_right, y_right, cur_depth + 1
                ),
                is_leaf=False,
            )
            self._nodes.append(node)
            return node

    def fit(self, X, y):
        self._root = self._generate_node(X, y, 0)

    def predict(self, X):
        def predict_single(datum):
            cur_node = self._root
            while not cur_node.is_leaf:
                cur_node = cur_node.get_child(datum)
            return cur_node.label

        if not self._root:
            raise Exception("Decision tree has not been trained.")
        size = X.shape[0]
        predictions = np.empty((size,), dtype=float)
        for i in range(size):
            predictions[i] = predict_single(X[i, :])
        return predictions

    def score(self, data, labels):
        if not self._root:
            raise Exception("Decision tree has not been trained.")
        predictions = self.predict(data)
        correct_count = np.count_nonzero(predictions == labels)
        return correct_count / labels.shape[0]
